import json
import os
import socket
import base64
import logging
import subprocess
import re
from pathlib import Path

import pyroute2
from pyroute2 import IPDB, WireGuard, NetlinkError
from nacl.public import PrivateKey

from platform_agent.cmd.lsmod import module_loaded
from platform_agent.cmd.wg_show import get_wg_listen_port
from platform_agent.files.tmp_files import get_peer_metadata
from platform_agent.lib.ctime import now
from platform_agent.routes import Routes
from platform_agent.wireguard.helpers import find_free_port, get_peer_info, WG_NAME_PATTERN

logger = logging.getLogger()


class WgConfException(Exception):
    pass


def delete_interface(ifname):
    subprocess.run(['ip', 'link', 'del', ifname], check=False, stderr=subprocess.DEVNULL)



def create_interface(ifname):
    try:
        subprocess.run(['ip', 'link', 'add', 'dev', ifname, 'type', 'wireguard'], check=True, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        pass


def set_interface_up(ifname):
    try:
        subprocess.run(['ip', 'link', 'set', 'up', ifname], check=True, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        pass


def set_interface_ip(ifname, ip):
    try:
        subprocess.run(['ip', 'address', 'add', 'dev', ifname, ip], check=True, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        pass


def add_iptable_rules(ips: list):
    for ip in ips:
        try:
            # Check if already exists, if not - create
            subprocess.run(['iptables', '-C', 'FORWARD', '-p', 'all', '-s', ip, '-j', 'ACCEPT'], check=True,
                           stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError:
            subprocess.run(
                ['iptables', '-A', 'FORWARD', '-p', 'all', '-s', ip, '-j', 'ACCEPT'],
                check=True, stderr=subprocess.DEVNULL
            )


def delete_iptable_rule(ips: list):
    for ip in ips:
        subprocess.run(
            ['iptables', '-D', 'FORWARD', '-p', 'all', '-s', ip, '-j', 'ACCEPT'],
            check=False,
            stderr=subprocess.DEVNULL
        )


class WgConf():

    def __init__(self, client=None):

        self.wg_kernel = module_loaded('wireguard')
        self.wg = WireGuard() if self.wg_kernel else WireguardGo()
        self.ipdb = IPDB()
        self.routes = Routes()
        self.client = client

    def create_syntropy_interfaces(self, ifaces):
        result = []
        for ifname in ifaces.keys():
            int_data = self.create_interface(ifname, ifaces[ifname].get('ip'))
            if int_data.get('public_key') != ifaces[ifname].get('public_key'):
                result.append(
                    {
                        "fn": "create_interface",
                        "data": int_data
                    }
                )

    @staticmethod
    def get_wg_interfaces():
        with IPDB() as ipdb:
            current_interfaces = [k for k, v in ipdb.by_name.items() if re.match(WG_NAME_PATTERN, k)]
        return current_interfaces

    def clear_interfaces(self, dump):
        remote_interfaces = [d['args']['ifname'] for d in dump if d['fn'] == 'create_interface']
        current_interfaces = self.get_wg_interfaces()
        remove_interfaces = set(current_interfaces) - set(remote_interfaces)
        logger.debug(
            f"Clearing interfaces REMOTE - {remote_interfaces}, CURRENT - {current_interfaces} REMOVE={remove_interfaces}"
        )
        for interface in remove_interfaces:
            self.remove_interface(interface)

    def clear_unused_routes(self, dump):
        remote_peers = [d['args'] for d in dump if d['fn'] == 'add_peer']
        remote_interfaces = [d['args']['ifname'] for d in dump if d['fn'] == 'create_interface']
        for ifname in remote_interfaces:
            allowed_ips = []
            remote_peers = [allowed_ips.extend(peer['allowed_ips']) for peer in remote_peers if peer and peer['ifname'] == ifname]
            self.routes.clear_unused_routes(ifname, allowed_ips)


    def clear_peers(self, dump):
        remote_peers = [d['args']['public_key'] for d in dump if d['fn'] == 'add_peer']
        current_interfaces = self.get_wg_interfaces()
        for iface in current_interfaces:
            peers = get_peer_info(iface, self.wg)
            for peer in peers:
                if peer not in remote_peers:
                    self.remove_peer(iface, peer)

    def get_wg_keys(self, ifname):
        private_key_path = f"/etc/syntropy-agent/privatekey-{ifname}"
        public_key_path = f"/etc/syntropy-agent/publickey-{ifname}"
        private_key = Path(private_key_path)
        public_key = Path(public_key_path)
        if not private_key.is_file() or not public_key.is_file():
            privKey = PrivateKey.generate()
            pubKey = base64.b64encode(bytes(privKey.public_key))
            privKey = base64.b64encode(bytes(privKey))
            base64_privKey = privKey.decode('ascii')
            base64_pubKey = pubKey.decode('ascii')
            private_key.write_text(base64_privKey)
            public_key.write_text(base64_pubKey)
            private_key.chmod(0o600)
            public_key.chmod(0o600)

        if self.wg_kernel:
            return public_key.read_text().strip(), private_key.read_text().strip()
        else:
            return public_key.read_text().strip(), private_key_path

    def next_free_port(self, port=1024, max_port=65535):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        while port <= max_port:
            try:
                sock.bind(('', port))
                sock.close()
                return port
            except OSError:
                port += 1
        raise IOError('no free ports')

    def create_interface(self, ifname, internal_ip, listen_port=None, **kwargs):
        public_key, private_key = self.get_wg_keys(ifname)
        peer_metadata = {'metadata': get_peer_metadata(public_key=public_key)}
        logger.info(
            f"[WG_CONF] - Creating interface {ifname}, {internal_ip} - wg_kernel={self.wg_kernel}",
            extra={'metadata': peer_metadata}
        )

        if self.wg_kernel:
            create_interface(ifname)
        else:
            self.wg.create_interface(ifname)
        set_interface_up(ifname)
        set_interface_ip(ifname, internal_ip)
        self.routes.clear_unused_iface_addrs(ifname, internal_ip.split('/')[0])

        if os.environ.get("SYNTROPY_PORT_RANGE") and not listen_port:
            listen_port = find_free_port()
        try:
            self.wg.set(
                ifname,
                private_key=private_key,
                listen_port=listen_port
            )
        except NetlinkError as error:
            if error.code != 98:
                raise
            else:
                # if port was taken before creating.
                self.wg.set(
                    ifname,
                    private_key=private_key,
                )
        listen_port = self.get_listening_port(ifname)
        if not listen_port:
            listen_port = find_free_port()
            self.wg.set(
                ifname,
                private_key=private_key,
                listen_port=listen_port
            )

        result = {
            "public_key": public_key,
            "listen_port": int(listen_port),
            "ifname": ifname
        }
        logger.debug(
            f"[WG_CONF] - interface_created {result}",
            extra={'metadata': peer_metadata}
        )
        return result

    def add_peer(self, ifname, public_key, allowed_ips, gw_ipv4, endpoint_ipv4=None, endpoint_port=None):
        peer_metadata = get_peer_metadata(public_key=public_key)
        if self.wg_kernel:
            try:
                peer_info = get_peer_info(ifname=ifname, wg=self.wg)
            except ValueError as e:
                raise WgConfException(str(e))
            old_ips = set(peer_info.get(public_key, [])) - set(allowed_ips)
            self.routes.ip_route_del(ifname, old_ips)
        peer = {'public_key': public_key,
                'endpoint_addr': endpoint_ipv4,
                'endpoint_port': endpoint_port,
                'persistent_keepalive': 15,
                'allowed_ips': allowed_ips}
        self.wg.set(ifname, peer=peer)
        statuses = self.routes.ip_route_add(ifname, allowed_ips, gw_ipv4)
        add_iptable_rules(allowed_ips)
        self.client.send_log(json.dumps({
            'id': "ID." + str(now()),
            'executed_at': now(),
            "type": "WG_ROUTE_STATUS",
            "data": {
                "connection_id": peer_metadata.get('connection_id'),
                "public_key": public_key,
                "statuses": statuses,
            }
        }))

    def remove_peer(self, ifname, public_key, allowed_ips=None):

        if ifname not in self.get_wg_interfaces():
            logger.debug(f'[WG_CONF] Remove peer - [{ifname}] does not exist')
            return

        peer = {
            'public_key': public_key,
            'remove': True
        }
        try:
            self.wg.set(ifname, peer=peer)
            if allowed_ips:
                self.routes.ip_route_del(ifname, allowed_ips)
                delete_iptable_rule(allowed_ips)
        except pyroute2.netlink.exceptions.NetlinkError as error:
            if error.code != 19:
                raise
        return

    def remove_interface(self, ifname):
        logger.debug(f'[WG_CONF] Removing interfcae - [{ifname}]')
        delete_interface(ifname)
        logger.debug(f'[WG_CONF] Removed interfcae - [{ifname}]')
        return

    def get_listening_port(self, ifname):
        if self.wg_kernel:
            wg_info = dict(self.wg.info(ifname)[0]['attrs'])
            return wg_info['WGDEVICE_A_LISTEN_PORT']

        else:
            wg_info = self.wg.info(ifname)
            return wg_info['listen_port']


class WireguardGo:

    def set(self, ifname, peer=None, private_key=None, listen_port=None):
        full_cmd = f"wg set {ifname}".split(' ')
        if peer:
            allowed_ips_cmd = ""
            if not peer.get('remove'):
                for ip in peer.get('allowed_ips', []):
                    allowed_ips_cmd += f"allowed-ips {ip} "
                peer_cmd = f"peer {peer['public_key']} {allowed_ips_cmd}endpoint {peer['endpoint_addr']}:{peer['endpoint_port']} persistent-keepalive 15".split(
                    ' ')
            else:
                peer_cmd = f"peer {peer['public_key']} remove".split(' ')
            full_cmd += peer_cmd
        if private_key:
            private_key_cmd = f"private-key {private_key}".split(' ')
            full_cmd += private_key_cmd
            if not listen_port:
                listen_port = find_free_port()
        if listen_port:
            listen_port_cmd = f"listen-port {listen_port}".split(' ')
            full_cmd += listen_port_cmd

        result_set = subprocess.run(full_cmd, encoding='utf-8', stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        complete_output = result_set.stdout or result_set.stderr
        complete_output = complete_output or 'Success'
        logger.debug(f"[Wireguard-go] - WG SET - {complete_output} , args {full_cmd}")
        return complete_output

    def create_interface(self, ifname):
        try:
            result_set = subprocess.Popen(
                ['wireguard-go', ifname],
                encoding='utf-8',
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                start_new_session=True,
            )
            result_set.wait(timeout=2)
        except FileNotFoundError:
            raise WgConfException(f'Wireguard-go missing')
        complete_output = result_set.stdout or result_set.stderr
        complete_output = complete_output or 'Success'
        logger.debug(f"[Wireguard-go] - WG Create - {complete_output.read()} , args {ifname}")
        return complete_output

    def info(self, ifname):
        return {
            "listen_port": get_wg_listen_port(ifname)
        }



import logging
import uuid
import json
import os

from logging.config import dictConfig
from pathlib import Path

from platform_agent.lib.ctime import now

logger = logging.getLogger()

class CustomLoggerAdapter(logging.LoggerAdapter):

    def process(self, msg, kwargs):
        """
        Process the Logging message and keyword arguments passed in to
        a logging call to insert contextual information. The extra argument
        of the LoggerAdapter will be merged with the extra argument of the
        logging call where the logging call's argument take precedence.
        """
        try:
            kwargs["extra"] = {**self.extra, **kwargs["extra"]}
        except KeyError as e:
            kwargs["extra"] = self.extra
        return msg, kwargs

class PublishLogToSessionHandler(logging.Handler):
    def __init__(self, session):
        logging.Handler.__init__(self)
        self.session = session
        self.log_id = str(uuid.uuid4())

    def emit(self, record):
        if not self.session.active:
            return
        metadata = getattr(record, "metadata", {})
        self.session.send_log(json.dumps({
            'id': self.log_id,
            'executed_at': now(),
            'type': 'LOGGER',
            'data': {'severity': record.levelname, 'message': record.getMessage(), 'metadata': metadata}
        }))


def configure_logger():

    log_path = "/var/log/syntropy-platform"

    log_file = Path(f"{log_path}/agent.log")

    syntropy_platform_dir = Path(f"{log_path}")

    if not syntropy_platform_dir.is_dir():
        syntropy_platform_dir.mkdir()
    if not log_file.is_file():
        log_file.write_text('')

    if os.environ.get('SYNTROPY_LOG_LEVEL', '20').isdigit():
        loglevel = int(os.environ.get('SYNTROPY_LOG_LEVEL', '20'))
    else:
        loglevel = os.environ.get('SYNTROPY_LOG_LEVEL', 'INFO')

    logging_config = dict(
        version=1,
        formatters={
            'f': {
                'format': '%(asctime)-24s %(levelname)-8s %(message)s'
            }
        },
        handlers={
            'h': {
                'class': 'logging.StreamHandler',
                'formatter': 'f',
                'level': loglevel
            },
            'file': {
                'level': 'DEBUG',
                'class': 'logging.handlers.RotatingFileHandler',
                'formatter': 'f',
                'filename': os.environ.get('SYNTROPY_LOG_FILE', "/var/log/syntropy-platform/agent.log")
            }
        },
        root={
            'handlers': ['h', 'file'],
            'level': loglevel,
        },
    )

    dictConfig(logging_config)

    logging.getLogger("pyroute2").setLevel(logging.ERROR)


#!/usr/bin/env python3
import iperf3

import logging
import threading
import time

logger = logging.getLogger()


class IperfServer(threading.Thread):

    def __init__(self):
        super().__init__()
        self.server = iperf3.Server()
        self.stop_iperf = threading.Event()
        self.daemon = True
        threading.Thread.__init__(self)

    @staticmethod
    def test_speed(hosts: list):
        now = int(time.time())
        results = {}
        client = iperf3.Client()
        for host in hosts:
            client.server_hostname = host
            result = client.run()
            data = [
                {
                    "time": now,
                    "host": host,
                    "fields": {
                        "upload": int(result.sent_Mbps),
                        "download": int(result.received_Mbps),
                        "retransmits": int(result.retransmits),
                    },
                }
            ]
            results[host] = data
            client.defaults()
        return results

    def run(self):
        while not self.stop_iperf.is_set():
            self.server.run()

    def join(self, timeout=None):
        self.stop_iperf.set()
        super().join(timeout)



import json

from platform_agent.config.settings import AGENT_PATH_TMP


def read_tmp_file(file_type='iface_info'):
    """Read iface file"""
    try:
        with open(f"{AGENT_PATH_TMP}/{file_type}") as json_file:
            rez = json_file.read()
            try:
                data = json.loads(rez)
            except json.JSONDecodeError:
                data = {}
    except FileNotFoundError:
        data = {}
    return data


def update_tmp_file(data, file_name):
    iface_info_path = f"{AGENT_PATH_TMP}/{file_name}"
    with open(iface_info_path, 'w+') as file:
        json.dump(data, file, indent=4)
        file.close()


def update_tmp_config_dump(cmd, file_name="config_dump"):
    data = read_tmp_file(file_name)
    cmds = data.get('vpn', [])
    cmds.append(cmd)
    data['vpn'] = cmds
    update_tmp_file(data, file_name)


def get_peer_metadata(file_name="config_dump", public_key=None, identifier='public_key'):
    data = read_tmp_file(file_name)
    cmds = data.get('vpn', [])
    peer_metadata = {}
    for cmd in cmds:
        if cmd['fn'] != 'add_peer':
            continue
        peer_metadata[cmd['args'][identifier]] = cmd['metadata']
    if public_key:
        return peer_metadata.get(public_key, {})
    return peer_metadata


def get_agent_id_by_text(text, file_name="config_dump"):
    data = read_tmp_file(file_name)
    cmds = data.get('vpn', [])
    for cmd in cmds:
        if cmd.get('metadata') and text in str(cmd):
            return cmd['metadata'].get('agent_id')
    return "UNKNOWN"



from platform_agent.files.tmp_files import get_peer_metadata

import mock


@mock.patch('platform_agent.files.tmp_files.read_tmp_file')
def test_config_info(patch_read_tmp_file, agent_dump, peer_file_read):
    patch_read_tmp_file.return_value = agent_dump
    peer_data = get_peer_metadata()
    assert peer_data == peer_file_read



from pytest import fixture


@fixture
def CONFIG_INFO():
    return 'CONFIG_INFO'


@fixture
def GET_INFO():
    return 'GET_INFO'


@fixture
def WG_INFO():
    return 'WG_INFO'


@fixture
def request_id():
    return 'TEST_01'


@fixture
def wg_int_name():
    return 'p2p_agent'


@fixture
def config_info_int_check():
    return {'agent_id': 71, 'vpn': [
        {'args': {'ifname': 'p2p_15_g8y3', 'internal_ip': '10.69.0.22/31'}, 'fn': 'create_interface'},
        {'args': {'ifname': 'mesh_111_v2t6', 'internal_ip': '10.69.0.35/31'}, 'fn': 'create_interface'},
        {'args': {'ifname': 'gw_113_d5yg', 'internal_ip': '10.69.0.56/31'}, 'fn': 'create_interface'},
        {'args': {'ifname': 'mesh_112_wnyt', 'internal_ip': '10.69.0.48/31'}, 'fn': 'create_interface'}
    ]}


@fixture
def gather_initial_info_payload():
    return {}


@fixture
def gather_initial_info_payload_bad():
    return "BAD_PAYLOAD"


@fixture
def wg_info_payload():
    return {'interval': 5}


@fixture
def mock_wg_show():
    return ('interface: mesh_8_qazm\n'
            '  public key: /c2PsNqVkbKJnvcdNeh4itIBpsZYNMLksUtXGIRgSDc=\n'
            '  private key: (hidden)\n'
            '  listening port: 43345\n'
            '\n'
            'peer: lh9VWZKS8Vu4b3QTJMuLAajvYJqO6GD9orMt8TQUWhE=\n'
            '  endpoint: 40.85.151.171:41152\n'
            '  allowed ips: 10.69.0.12/32, 192.168.151.0/24\n'
            '  transfer: 0 B received, 2.79 MiB sent\n'
            '  persistent keepalive: every 15 seconds\n'
            '\n'
            'peer: RNlLq4YM2jXwZ3r9uAXOBf+zbYrqImndeLjAqIZLRGI=\n'
            '  endpoint: 140.238.215.51:51783\n'
            '  allowed ips: 10.69.0.11/32\n'
            '  transfer: 0 B received, 2.79 MiB sent\n'
            '  persistent keepalive: every 15 seconds\n'
            '\n'
            'peer: qlkykBhjGpuJkyayQJMw5fyLZyTGFfRTTnuXQLBhPQQ=\n'
            '  endpoint: 13.93.72.56:33632\n'
            '  allowed ips: 10.69.0.13/32, 172.17.0.0/16, 192.168.152.0/24, '
            '192.168.154.0/24\n'
            '  transfer: 0 B received, 2.79 MiB sent\n'
            '  persistent keepalive: every 15 seconds\n')


@fixture
def wg_show_dict():
    return [{'interface': 'mesh_8_qazm',
             'listening_port': '43345',
             'peers': [{'allowed_ips': ['10.69.0.12/32', '192.168.151.0/24'],
                        'endpoint': '40.85.151.171:41152',
                        'latest_handshake': None,
                        'peer': 'lh9VWZKS8Vu4b3QTJMuLAajvYJqO6GD9orMt8TQUWhE=',
                        'persistent_keepalive': 'every 15 seconds',
                        'preshared_key': None,
                        'transfer': '0 B received, 2.79 MiB sent'},
                       {'allowed_ips': ['10.69.0.11/32'],
                        'endpoint': '140.238.215.51:51783',
                        'latest_handshake': None,
                        'peer': 'RNlLq4YM2jXwZ3r9uAXOBf+zbYrqImndeLjAqIZLRGI=',
                        'persistent_keepalive': 'every 15 seconds',
                        'preshared_key': None,
                        'transfer': '0 B received, 2.79 MiB sent'},
                       {'allowed_ips': ['10.69.0.13/32',
                                        '172.17.0.0/16',
                                        '192.168.152.0/24',
                                        '192.168.154.0/24'],
                        'endpoint': '13.93.72.56:33632',
                        'latest_handshake': None,
                        'peer': 'qlkykBhjGpuJkyayQJMw5fyLZyTGFfRTTnuXQLBhPQQ=',
                        'persistent_keepalive': 'every 15 seconds',
                        'preshared_key': None,
                        'transfer': '0 B received, 2.79 MiB sent'}],
             'private_key': '(hidden)',
             'public_key': '/c2PsNqVkbKJnvcdNeh4itIBpsZYNMLksUtXGIRgSDc='}]


@fixture
def agent_dump():
    return \
        {
            "agent_id": 30,
            "vpn": [
                {
                    "fn": "create_interface",
                    "args": {
                        "ifname": "pmesh_51_eg0r",
                        "internal_ip": "10.69.0.10/32",
                        "listen_port": 57950
                    }
                },
                {
                    "fn": "add_peer",
                    "args": {
                        "allowed_ips": [
                            "10.69.0.11/32",
                            "192.168.151.0/24"
                        ],
                        "endpoint_ipv4": "23.99.200.18",
                        "endpoint_port": 50848,
                        "ifname": "pmesh_51_eg0r",
                        "public_key": "EhFJfT0UelKZRoYITH3NHyZmCPmvO85n7Fmw2wuwTBI=",
                        "gw_ipv4": "10.69.0.10"
                    },
                    "metadata": {
                        "device_id": "1123",
                        "device_name": "device_name",
                        "device_public_ipv4": "72.5.6.4"
                    }
                },
                {
                    "fn": "create_interface",
                    "args": {
                        "ifname": "smesh_51_eyus",
                        "internal_ip": "10.69.0.12/32",
                        "listen_port": 48849
                    }
                },
                {
                    "fn": "add_peer",
                    "args": {
                        "allowed_ips": [
                            "10.69.0.13/32",
                            "192.168.151.0/24"
                        ],
                        "endpoint_ipv4": "159.8.21.106",
                        "endpoint_port": 15566,
                        "ifname": "smesh_51_eyus",
                        "public_key": "EX19IgQzdPkM1tsfvSv+XOhhQ8JD/Q13e+i/H26K53U=",
                        "gw_ipv4": "10.69.0.12"
                    },
                    "metadata": {
                        "device_id": "9993",
                        "device_name": "9993-agent",
                        "device_public_ipv4": "73.5.6.4"
                    }
                }
            ]
        }


@fixture
def peer_file_read():
    return {
        "EhFJfT0UelKZRoYITH3NHyZmCPmvO85n7Fmw2wuwTBI=": {
            "device_id": "1123",
            "device_name": "device_name",
            "device_public_ipv4": "72.5.6.4"
        },
        "EX19IgQzdPkM1tsfvSv+XOhhQ8JD/Q13e+i/H26K53U=": {
            "device_id": "9993",
            "device_name": "9993-agent",
            "device_public_ipv4": "73.5.6.4"
        }
    }



import logging
import threading
import time
import json

import pyroute2

from platform_agent.config.settings import AGENT_PATH_TMP
from platform_agent.files.tmp_files import get_peer_metadata

logger = logging.getLogger()


def read_tmp_file(file_type='iface_info'):
    """Read iface file"""
    try:
        with open(f"{AGENT_PATH_TMP}/{file_type}") as json_file:
            rez = json_file.read()
            try:
                data = json.loads(rez)
            except json.JSONDecodeError:
                data = {}
    except FileNotFoundError:
        data = {}
    return data


class InterfaceWatcher(threading.Thread):

    def __init__(self):
        super().__init__()
        self.iface_watcher = threading.Event()
        self.watcher = threading.Event()
        self.daemon = True

    def update_iface_info_file(self, data):
        iface_info_path = f"{AGENT_PATH_TMP}/iface_info"
        with open(iface_info_path, 'w+') as iface_info_file:
            json.dump(data, iface_info_file)
            iface_info_file.close()

    def run(self):
        with pyroute2.IPDB() as ipdb:
            while not self.iface_watcher.is_set():
                    peers_metadata = get_peer_metadata(identifier='ifname')
                    res = {k: v for k, v in ipdb.by_name.items()}
                    payload = {}
                    for ifname in res.keys():
                        if not res[ifname].get('ipaddr') and not res[ifname]['ipaddr'].ipv4 or not len(res[ifname]['ipaddr'].ipv4):
                            continue
                        internal_ip = f"{res[ifname]['ipaddr'].ipv4[0]['address']}/{res[ifname]['ipaddr'].ipv4[0]['prefixlen']}"
                        payload[ifname] = {
                            'internal_ip': internal_ip,
                            'kind': res[ifname]['kind'],
                            'metadata': peers_metadata.get(ifname, {})
                        }
                    self.update_iface_info_file(payload)
                    time.sleep(1)

    def join(self, timeout=None):
        self.watcher.set()
        super().join(timeout)



import json
import logging
import threading
import time
import psutil

from platform_agent.config.settings import Config
from platform_agent.lib.ctime import now
from pyroute2 import IPDB
logger = logging.getLogger()


class DummyNetworkWatcher(threading.Thread):

    def __init__(self, ws_client):
        super().__init__()
        self.ws_client = ws_client
        self.stop_network_watcher = threading.Event()
        with IPDB() as ipdb:
            self.ifaces = [k for k, v in ipdb.by_name.items() if any(
                substring in k for substring in ['syntropy_'])]
        self.daemon = True

    def run(self):
        ex_result = []
        with IPDB() as ipdb:
            while not self.stop_network_watcher.is_set():
                udp = psutil.net_connections(kind='udp')
                udp_info = [{x.laddr.ip: x.laddr.port} for x in udp]
                tcp = psutil.net_connections(kind='tcp')
                tcp_info = [{x.laddr.ip: x.laddr.port} for x in tcp]
                result = []
                for iface in self.ifaces:
                    intf = ipdb.interfaces[iface]
                    for k, v in dict(intf['ipaddr']).items():
                        udp_ports = [ip[k] for ip in udp_info if ip.get(k)]
                        tcp_ports = [ip[k] for ip in tcp_info if ip.get(k)]
                        result.append(
                            {
                                'agent_network_subnets': [f"{k}/{v}"],
                                'agent_network_iface': iface,
                                'agent_network_ports': {'udp': udp_ports, 'tcp': tcp_ports},
                            }
                        )
                status = getattr(self.ws_client.ws, 'sock')
                result.extend(Config.get_valid_allowed_ips())
                if result != ex_result and status and status.status:
                    self.ws_client.send(json.dumps({
                        'id': "ID." + str(time.time()),
                        'executed_at': now(),
                        'type': 'HW_SERVICE_INFO',
                        'data': result
                    }))
                    ex_result = result
                time.sleep(3)

    def join(self, timeout=None):
        self.stop_network_watcher.set()
        super().join(timeout)


import logging
import os
import json
import socket
import queue
import threading
import traceback
import time

import requests
import websocket

from platform_agent.lib.ctime import now
from platform_agent.agent_api import AgentApi
from platform_agent.config.logger import PublishLogToSessionHandler
from platform_agent.wireguard.helpers import check_if_wireguard_installled
from platform_agent.__main__ import __version__

logger = logging.getLogger()


class AgentRunner:

    STOP_MESSAGE = now()

    def __init__(self, ws):
        self.ws = ws
        self.queue = queue.Queue()
        self.active = None
        self.agent_api = AgentApi(self)

        logging.root.addHandler(PublishLogToSessionHandler(self))

    def run(self):
        while True:
            message = self.queue.get()
            if message == self.STOP_MESSAGE:
                break
            request = json.loads(message)
            logger.debug(f"[RUNNER] Parsed request | {request}")
            try:
                result = self.agent_api.call(request['type'], request['data'], request['id'])
            except:  # noqa
                # Catch all exceptions that not handled
                traceback.print_exc()
                result = {
                    'error': {
                        'traceback': traceback.format_exc(),
                        'payload': request
                    }
                }
                logger.error(result)
            self.queue.task_done()
            logger.debug(f"[RUNNER] Result | {result}")
            if result:
                payload = self.create_response(request, result)
                self.send(payload)

    @staticmethod
    def create_response(request, result):
        payload = {
            'id': request['id'],
            'executed_at': now(),
            'type': request['type'],
        }
        if isinstance(result, dict) and ('error' in result):
            payload.update(result)
        else:
            payload.update({'data': result})
        return json.dumps(payload)

    def send(self, message):
        status = getattr(self.ws, 'sock')
        if status and status.status:
            logger.debug(f"[SENDING]: {message}")
            try:
                self.ws.send(message)
            except:
                pass
        else:
            logger.error("[SENDING]: websocket offline")

    def send_log(self, message):
        status = getattr(self.ws, 'sock')
        if status and status.status:
            try:
                self.ws.send(message)
            except:
                pass


class WebSocketClient(threading.Thread):

    def __init__(self, host, api_key, ssl="wss"):
        threading.Thread.__init__(self)

        if check_if_wireguard_installled():
            status = 'OK'
        else:
            status = 'WG_ERROR'

        self.host = host
        self.active = True
        websocket.enableTrace(False)
        self.connection_url = f"{ssl}://{host}"
        self.ws = websocket.WebSocketApp(
            self.connection_url,
            header={
                'authorization': api_key,
                'x-deviceid': self.generate_device_id(),
                'x-devicename': os.environ.get('SYNTROPY_AGENT_NAME', socket.gethostname()),
                'x-devicestatus': status,
                'x-agentversion': __version__,
            },
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
            on_open=self.on_open
        )
        self.agent_runner = AgentRunner(self.ws)
        threading.Thread(target=self.agent_runner.run).start()
        self.ws.on_message = self.on_message
        self.ws.on_open = self.on_open

    def run(self):
        while True and self.active:
            logger.debug(f"[AGENT-{__version__}] Connecting {self.connection_url}")
            self.ws.run_forever(ping_interval=60, ping_timeout=10)
            logger.warning(f"[AGENT-{__version__}] Disconnected {self.connection_url}")
            time.sleep(10)

    def on_message(self, message):
        logger.debug(f"[WEBSOCKET] Received | {message}")
        logger.debug(f"[WEBSOCKET] Queue size | {self.agent_runner.queue.qsize()}")
        self.agent_runner.queue.put(message)

    def on_error(self, error):
        self.agent_runner.active = False
        logger.error(f"[WEBSOCKET] Error | {error}")
        self.ws.close()
        time.sleep(10)

    def on_close(self):
        self.agent_runner.active = False
        logger.debug("[WEBSOCKET] Close")
        time.sleep(10)

    def on_open(self):
        logger.debug("[WEBSOCKET] Connection open")
        self.agent_runner.active = True

    def stop(self):
        self.ws.close()
        self.agent_runner.active = False
        self.agent_runner.queue.put(self.agent_runner.STOP_MESSAGE)

    @staticmethod
    def getserial():
        # Extract serial from cpuinfo file
        cpuserial = "0000000000000000"
        f = open('/proc/cpuinfo', 'r')
        for line in f:
            if line[0:6] == 'Serial':
                cpuserial = line[10:26]
        f.close()

        return cpuserial

    def generate_device_id(self):
        try:
            with open('/sys/class/dmi/id/product_uuid', 'r') as file:
                machine_id = file.read().replace('\n', '')
        except FileNotFoundError:
            try:
                with open('/etc/machine-id', 'r') as file:
                    machine_id = file.read().replace('\n', '') + requests.get("https://ip.syntropystack.com/").json()
            except FileNotFoundError:
                machine_id = self.getserial()

        return machine_id




import json
import logging
import threading
import time

from icmplib import multiping

from platform_agent.lib.ctime import now

logger = logging.getLogger()


class AutopingClient(threading.Thread):

    def __init__(self, client, ips, interval, response_limit=5):
        super().__init__()
        self.client = client
        self.interval = interval
        self.response_limit = response_limit
        self.hosts = ips
        self.stop_autoping = threading.Event()
        self.daemon = True
        threading.Thread.__init__(self)

    def run(self):
        while not self.stop_autoping.is_set():
            pings = []
            ping_res = multiping(self.hosts, count=5, interval=0.5, max_threads=2)
            ping_res.sort(key=lambda x: x.avg_rtt)
            for res in ping_res:
                if res.is_alive:
                    pings.append({
                        "ip": res.address,
                        "latency_ms": res.avg_rtt if res.is_alive else -1,
                        "packet_loss": res.packet_loss if res.is_alive else 1
                    })
                if len(pings) >= self.response_limit:
                    break

            self.client.send_log(json.dumps({
                'id': "ID." + str(time.time()),
                'executed_at': now(),
                'type': 'AUTO_PING',
                'data': {"pings": pings}
            }))
            time.sleep(int(self.interval))

    def join(self, timeout=None):
        self.stop_autoping.set()
        super().join(timeout)



from platform_agent.wireguard import WgConf

import mock


@mock.patch('platform_agent.agent_api.WgConf.remove_interface')
@mock.patch('platform_agent.agent_api.WgConf.get_wg_interfaces')
def test_config_info(patch_get_wg_interfaces, patch_remove_interface, config_info_int_check, wg_int_name):
    patch_get_wg_interfaces.return_value = [wg_int_name]
    wgconf = WgConf()
    wgconf.clear_interfaces(config_info_int_check['vpn'])
    assert patch_remove_interface.call_args.args[0] == wg_int_name



import re
import os
import json

import dataclasses
from typing import List


class DataclassJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)


@dataclasses.dataclass
class WgPeer:
    peer: str
    allowed_ips: List[str]
    preshared_key: str = None
    endpoint: str = None
    latest_handshake: str = None
    persistent_keepalive: str = None
    transfer: str = None


@dataclasses.dataclass
class WgInterface:
    interface: str
    listening_port: str = None
    private_key: str = None
    public_key: str = None
    peers: List[WgPeer] = dataclasses.field(default_factory=[])


class WireGuardRead:
    def __init__(self):
        self.interface_regex = r'((?:[^\n][\n]?)+)'
        self.parameter_regex = r'(^.+): (.+$)'
        self.stdin = None

    def wg_info(self, ifname=None):
        if ifname:
            grep = f" {ifname}"
        else:
            grep = ""
        self.stdin = os.popen('wg show' + grep).read()
        output = []
        for i in map(self.make_json, self.all_interfaces()):
            if 'interface' in i:
                interface = WgInterface(peers=[], **i)
                output.append(interface)
            else:
                interface.peers.append(WgPeer(**i))
        output = json.loads(json.dumps(output, cls=DataclassJSONEncoder))
        return output

    def all_interfaces(self):
        interfaces = re.findall(self.interface_regex, self.stdin, re.MULTILINE)
        if interfaces:
            return interfaces
        else:
            return []

    def format_value(self, key, value):
        key = self.format_key(key)
        array_format = ['allowed_ips']

        if key in array_format:
            value = value.split(', ')

        return value

    @staticmethod
    def format_key(key):
        return key.replace('  ', '').replace(' ', '_')

    def make_json(self, s):
        confing = re.findall(self.parameter_regex, s, re.MULTILINE)
        return {self.format_key(key): self.format_value(key, value) for key, value in confing}



import time
import json
import threading
import re

from platform_agent.files.tmp_files import read_tmp_file
from platform_agent.wireguard.helpers import WG_NAME_PATTERN
from platform_agent.lib.ctime import now


class BWDataCollect(threading.Thread):

    def __init__(self, client, interval=10):
        super().__init__()
        self.interval = interval
        self.client = client
        self.stop_BWDataCollect = threading.Event()
        self.daemon = True

    @staticmethod
    def get_int_info(t, iface):
        try:
            with open('/sys/class/net/' + iface + '/statistics/' + t, 'r') as f:
                data = f.read()
                return int(data)
        except FileNotFoundError:
            return -1

    @staticmethod
    def get_iface_info_set(iface, interval=10):
        tx_bytes = BWDataCollect.get_int_info('tx_bytes', iface)
        rx_bytes = BWDataCollect.get_int_info('rx_bytes', iface)

        tx_dropped = BWDataCollect.get_int_info('tx_dropped', iface)
        tx_errors = BWDataCollect.get_int_info('tx_errors', iface)
        tx_packets = BWDataCollect.get_int_info('tx_packets', iface)

        rx_dropped = BWDataCollect.get_int_info('rx_dropped', iface)
        rx_errors = BWDataCollect.get_int_info('rx_errors', iface)
        rx_packets = BWDataCollect.get_int_info('rx_packets', iface)

        time.sleep(interval)

        tx_bytes_after = BWDataCollect.get_int_info('tx_bytes', iface)
        rx_bytes_after = BWDataCollect.get_int_info('rx_bytes', iface)

        tx_dropped_after = BWDataCollect.get_int_info('tx_dropped', iface)
        tx_errors_after = BWDataCollect.get_int_info('tx_errors', iface)
        tx_packets_after = BWDataCollect.get_int_info('tx_packets', iface)

        rx_dropped_after = BWDataCollect.get_int_info('rx_dropped', iface)
        rx_errors_after = BWDataCollect.get_int_info('rx_errors', iface)
        rx_packets_after = BWDataCollect.get_int_info('rx_packets', iface)

        tx_speed_mbps = round((tx_bytes_after - tx_bytes) / 10000000.0, 4)
        rx_speed_mbps = round((rx_bytes_after - rx_bytes) / 10000000.0, 4)
        tx_dropped = (tx_dropped_after - tx_dropped)
        tx_errors = (tx_errors_after - tx_errors)
        tx_packets = (tx_packets_after - tx_packets)
        rx_dropped = (rx_dropped_after - rx_dropped)
        rx_errors = (rx_errors_after - rx_errors)
        rx_packets = (rx_packets_after - rx_packets)
        return {
            'iface': iface,
            'tx_speed_mbps': tx_speed_mbps,
            'rx_speed_mbps': rx_speed_mbps,
            'tx_dropped': tx_dropped,
            'tx_errors': tx_errors,
            'tx_packets': tx_packets,
            'rx_dropped': rx_dropped,
            'rx_errors': rx_errors,
            'rx_packets': rx_packets,
            'interval': interval,
        }

    def run(self):
        while not self.stop_BWDataCollect.is_set():
            interfaces = read_tmp_file(file_type='iface_info')
            wg_ifaces = {k: v for k, v in interfaces.items() if re.match(WG_NAME_PATTERN, k)}
            if not wg_ifaces:
                time.sleep(1)
            for iface in wg_ifaces:
                try:
                    result = [self.get_iface_info_set(iface, self.interval)]
                except FileNotFoundError:
                    continue
                if not result:
                    continue
                self.client.send_log(json.dumps({
                    'id': "UNKNOWN",
                    'executed_at': now(),
                    'type': 'IFACES_BW_DATA',
                    'data': result
                }))
                time.sleep(int(self.interval))

    def join(self, timeout=None):
        self.stop_BWDataCollect.set()
        super().join(timeout)


import docker


def format_networks_result(networks):
    result = []
    for network in networks:
        subnets = []
        for subnet in network['IPAM']['Config']:
            subnets.append(subnet['Subnet'])
        if subnets:
            result.append(
                {
                    'agent_network_id': network['Id'],
                    'agent_network_name': network.get('Name'),
                    'agent_network_subnets': subnets,
                }
            )
    return result


def format_container_result(containers):
    docker_client = docker.from_env()
    networks = docker_client.networks()
    conts = {}
    for network in networks:
        for k, v in network['Containers'].items():
            if not v['IPv4Address']:
                continue
            if conts.get(k) and conts[k]['IPv4Address'] != v['IPv4Address']:
                conts[k]['IPv4Address'].append(v['IPv4Address'].split('/')[0])
                conts[k]['network_names'].append(network['Name'])
                continue
            conts[k] = v
            conts[k]['IPv4Address'] = [conts[k].get('IPv4Address').split('/')[0]]
            conts[k]['network_names'] = [network.get('Name')]
    result = []
    for container in containers:
        ports = {'udp': [], 'tcp': []}
        for port in container.get('Ports', []):
            private_port = port.get('PrivatePort')
            public_port = port.get('PublicPort')
            port_type = port.get('Type')
            if not port_type:
                continue
            if private_port and private_port not in ports[port_type]:
                ports[port_type].append(private_port)
            if public_port and public_port not in ports[port_type]:
                ports[port_type].append(public_port)
        container_info = conts.get(container['Id'])

        if container_info:

            container_conf = docker_client.inspect_container(container['Id'])['Config']

            try:
                container_info['name'] = \
                    [name for name in container_conf.get('Env', []) if 'SYNTROPY_SERVICE_NAME' in name][0].split('=')[1]
            except IndexError:
                container_info['name'] = container_conf.get('Domainname')
                if not container_info['name']:
                    container_info['name'] = container_info.get('Name')

            result.append(
                {
                    'agent_container_id': container['Id'],
                    'agent_container_name': container_info.get('name'),
                    'agent_container_ips': container_info.get('IPv4Address'),
                    'agent_container_networks': container_info.get('network_names'),
                    'agent_container_ports': ports,
                    'agent_container_state': container.get('State'),
                }
            )
    return result



import json
import logging
import threading
import os

from platform_agent.lib.ctime import now
from platform_agent.cmd.lsmod import module_loaded
from platform_agent.files.tmp_files import update_tmp_file
from platform_agent.lib.get_info import gather_initial_info
from platform_agent.network.exporter import NetworkExporter
from platform_agent.network.kubernetes_watcher import KubernetesNetworkWatcher
from platform_agent.wireguard import WgConfException, WgConf, WireguardPeerWatcher
from platform_agent.docker_api.docker_api import DockerNetworkWatcher
from platform_agent.network.dummy_watcher import DummyNetworkWatcher
from platform_agent.executors.wg_exec import WgExecutor
from platform_agent.network.network_info import BWDataCollect
from platform_agent.network.autoping import AutopingClient
from platform_agent.network.iperf import IperfServer
from platform_agent.network.iface_watcher import InterfaceWatcher
from platform_agent.rerouting.rerouting import Rerouting

logger = logging.getLogger()


class AgentApi:

    def __init__(self, runner, prod_mode=True):
        self.runner = runner
        self.wg_peers = None
        self.autoping = None
        self.wgconf = WgConf(self.runner)
        self.wg_executor = WgExecutor(self.runner)
        self.bw_data_collector = BWDataCollect(self.runner)
        if prod_mode:
            threading.Thread(target=self.wg_executor.run).start()
            threading.Thread(target=self.bw_data_collector.run).start()
            self.network_exporter = NetworkExporter().start()
            self.wg_peers = WireguardPeerWatcher(self.runner).start()
            self.interface_watcher = InterfaceWatcher().start()
            if module_loaded("wireguard"):
                os.environ["SYNTROPY_WIREGUARD"] = "true"
            if os.environ.get("SYNTROPY_NETWORK_API", '').lower() == "docker" and prod_mode:
                self.network_watcher = DockerNetworkWatcher(self.runner).start()
            if os.environ.get("SYNTROPY_NETWORK_API", '').lower() == "host" and prod_mode:
                self.network_watcher = DummyNetworkWatcher(self.runner).start()
            if os.environ.get("SYNTROPY_NETWORK_API", '').lower() == "kubernetes" and prod_mode:
                self.network_watcher = KubernetesNetworkWatcher(self.runner).start()
            self.rerouting = Rerouting(self.runner).start()

    def call(self, type, data, request_id):
        result = None
        try:
            if hasattr(self, type):
                logger.debug(f"[AGENT_API] Calling agent api {data}")
                if not isinstance(data, (dict, list)):
                    logger.error('[AGENT_API] data should be "DICT" type')
                    result = {'error': "BAD REQUEST"}
                else:
                    fn = getattr(self, type)
                    result = fn(data, request_id=request_id)
        except AttributeError as error:
            logger.warning(error)
            result = {'error': str(error)}
        return result

    def GET_INFO(self, data, **kwargs):
        return gather_initial_info(**data)

    def WG_INFO(self, data, **kwargs):
        if self.wg_peers:
            self.wg_peers.join(timeout=1)
            self.wg_peers = None
        self.wg_peers = WireguardPeerWatcher(self.runner, **data)
        self.wg_peers.start()

    def WG_CONF(self, data, **kwargs):
        self.wg_executor.queue.put({"data": data, "request_id": kwargs['request_id']})
        return False

    def AUTO_PING(self, data, **kwargs):
        if self.autoping:
            self.autoping.join(timeout=1)
            self.autoping = None
        self.autoping = AutopingClient(self.runner, **data)
        self.autoping.start()
        return False

    def CONFIG_INFO(self, data, **kwargs):
        update_tmp_file(data, 'config_dump')
        self.wgconf.clear_interfaces(data.get('vpn', []))
        self.wgconf.clear_peers(data.get('vpn', []))
        self.wgconf.clear_unused_routes(data.get('vpn', []))
        response = []
        for vpn_cmd in data.get('vpn', []):
            try:
                fn = getattr(self.wgconf, vpn_cmd['fn'])
                result = fn(**vpn_cmd['args'])
                if vpn_cmd['fn'] == 'create_interface' and result and\
                        (vpn_cmd['args'].get('public_key') != result.get('public_key') or
                         vpn_cmd['args'].get('listen_port') != result.get('listen_port')):
                    response.append({'fn': vpn_cmd['fn'], 'data': result})
            except WgConfException as e:
                logger.error(f"[CONFIG_INFO] [{str(e)}]")
        self.runner.send(json.dumps({
            'id': kwargs['request_id'],
            'executed_at': now(),
            'type': 'UPDATE_AGENT_CONFIG',
            'data': response
        }))

    def IPERF_SERVER(self, data, **kwargs):
        if self.iperf and data.get('status') == 'off':
            self.iperf.join(timeout=1)
            self.iperf = None
            return 'ok'
        if data.get('status'):
            self.iperf = IperfServer()
            IperfServer.start(self.runner)
            return 'ok'

    def IPERF_TEST(self, data, **kwargs):
        if data.get('hosts') and isinstance(data['hosts'], list):
            result = IperfServer.test_speed(**data)
            return result
        else:
            return {"error": "must be list"}


import threading
import time
import logging
import pyroute2
import ipaddress
import json
import re

from pyroute2 import WireGuard

from platform_agent.cmd.lsmod import module_loaded
from platform_agent.cmd.wg_info import WireGuardRead
from platform_agent.files.tmp_files import read_tmp_file
from platform_agent.routes import Routes
from platform_agent.lib.ctime import now

from platform_agent.wireguard.helpers import WG_NAME_PATTERN, ping_internal_ips, get_peer_info_all

logger = logging.getLogger()


def get_routing_info(wg):
    routing_info = {}
    peers_internal_ips = []
    interfaces = read_tmp_file(file_type='iface_info')
    res = {k: v for k, v in interfaces.items() if re.match(WG_NAME_PATTERN, k)}
    for ifname in res.keys():
        if not res[ifname].get('internal_ip'):
            continue
        internal_ip = res[ifname]['internal_ip']
        metadata = res[ifname]['metadata']
        peers = get_peer_info_all(ifname, wg, kind=res[ifname]['kind'])
        for peer in peers:
            try:
                peer_internal_ip = next(
                    (
                        ip for ip in peer['allowed_ips']
                        if
                        ipaddress.ip_address(ip.split('/')[0]) in ipaddress.ip_network(f"{internal_ip.split('/')[0]}/24",
                                                                                       False)
                    ),
                    None
                )
            except ValueError:
                continue
            if not peer_internal_ip:
                continue
            peers_internal_ips.append(peer_internal_ip.split('/')[0])
            peer['allowed_ips'].remove(peer_internal_ip)
            for allowed_ip in peer['allowed_ips']:
                if not routing_info.get(allowed_ip):
                    routing_info[allowed_ip] = {'ifaces': {}}
                routing_info[allowed_ip]['ifaces'][ifname] = {
                    'internal_ip': peer_internal_ip,
                    'metadata': metadata
                }
    return routing_info, peers_internal_ips


def get_interface_internal_ip(ifname):
    with pyroute2.IPDB() as ipdb:
        internal_ip = f"{ipdb.interfaces[ifname]['ipaddr'][0]['address']}"
        return internal_ip


def get_fastest_routes(wg):
    result = {}
    routing_info, peers_internal_ips = get_routing_info(wg)
    ping_results = ping_internal_ips(peers_internal_ips, icmp_id=20000)
    for dest, routes in routing_info.items():
        best_route = None
        best_ping = 9999
        for iface, data in routes['ifaces'].items():
            int_ip = data['internal_ip'].split('/')[0]
            if ping_results[int_ip]['latency_ms'] < best_ping:
                best_route = {'iface': iface, 'gw': data['internal_ip'], 'metadata': data.get('metadata')}
                best_ping = ping_results[int_ip]['latency_ms']
        result[dest] = best_route
    return result, ping_results


class Rerouting(threading.Thread):

    def __init__(self, client, interval=1):
        logger.debug(f"[REROUTING] Initializing")
        super().__init__()
        self.interval = interval
        self.client = client
        self.wg = WireGuard() if module_loaded("wireguard") else WireGuardRead()
        self.routes = Routes()
        self.stop_rerouting = threading.Event()
        self.daemon = True

    def run(self):
        logger.debug(f"[REROUTING] Running")
        previous_routes = {}
        while not self.stop_rerouting.is_set():
            new_routes, ping_data = get_fastest_routes(self.wg)
            for dest, best_route in new_routes.items():
                if not best_route or previous_routes.get(dest) == best_route:
                    continue
                # Do rerouting logic with best_route
                logger.debug(f"[REROUTING] Rerouting {dest} via {best_route}", extra={'metadata': best_route.get('metadata')})
                try:
                    self.routes.ip_route_replace(
                        ifname=best_route['iface'], ip_list=[dest],
                        gw_ipv4=get_interface_internal_ip(best_route['iface'])
                    )
                except KeyError:  # catch if interface was deleted while executing this code
                    continue
            previous_routes = new_routes
            time.sleep(int(self.interval))

    def send_latency_data(self, data):
        self.client.send_log(json.dumps({
            'id': "ID." + str(time.time()),
            'executed_at': now(),
            'type': 'PEERS_LATENCY_DATA',
            'data': data
        }))

    def join(self, timeout=None):
        self.stop_rerouting.set()
        super().join(timeout)



import json
import logging
import threading
import time

from pyroute2 import WireGuard

from platform_agent.cmd.lsmod import module_loaded
from platform_agent.lib.ctime import now
from platform_agent.wireguard.helpers import merged_peer_info
from platform_agent.cmd.wg_info import WireGuardRead


logger = logging.getLogger()


class WireguardPeerWatcher(threading.Thread):

    def __init__(self, client, interval=60):
        super().__init__()
        self.client = client
        self.interval = interval
        self.wg = WireGuard() if module_loaded("wireguard") else WireGuardRead()
        self.stop_peer_watcher = threading.Event()
        self.daemon = True

    def run(self):
        while not self.stop_peer_watcher.is_set():
            peer_info = merged_peer_info(self.wg)
            if not peer_info:
                time.sleep(1)
                continue
            self.client.send_log(json.dumps({
                'id': "UNKNOWN",
                'executed_at': now(),
                'type': 'IFACES_PEERS_BW_DATA',
                'data': peer_info
            }))
            time.sleep(int(self.interval))

    def join(self, timeout=None):
        self.stop_peer_watcher.set()
        super().join(timeout)



import datetime
import os
import ipaddress
import re
import psutil
import socket
from random import randint

from icmplib import multiping
from pyroute2 import NetlinkError

from platform_agent.cmd.lsmod import module_loaded, is_tool
from platform_agent.cmd.wg_info import WireGuardRead
from platform_agent.network.iface_watcher import read_tmp_file

WG_NAME_PATTERN = '[0-9]{10}(s1|s2|s3|p0)+(g|m|p)[Nn][Oo]'


def get_connection_status(latency_ms, packet_loss):
    if packet_loss >= 1:
        res = {'status': 'OFFLINE', 'status_reason': 'Packet loss 100%'}
    elif 0.01 <= packet_loss <= 1:
        res = {'status': 'WARNING', 'status_reason': 'Packet loss higher than 1%'}
    elif latency_ms >= 1000:
        res = {'status': 'WARNING', 'status_reason': 'Latency higher than 1000ms'}
    else:
        res = {'status': 'CONNECTED'}
    res.update(
        {
            "latency_ms": latency_ms,
            "packet_loss": packet_loss,
        }
    )
    return res


def find_free_port():
    ports_start = 49152
    ports_end = 65535
    if os.environ.get("SYNTROPY_PORT_RANGE"):
        try:
            ports = os.environ["SYNTROPY_PORT_RANGE"]
            ports = ports.split('-')
            ports_start = int(ports[0])
            ports_end = int(ports[1])
        except (IndexError, ValueError):
            pass
    port = randint(ports_start, ports_end)
    portsinuse = []
    while True:
        conns = psutil.net_connections()
        for conn in conns:
            portsinuse.append(conn.laddr[1])
        if port in portsinuse:
            port = randint(ports_start, ports_end)
        else:
            break
    return port


def get_iface_public_key(ifname):
    wg = WireGuardRead()
    ifaces = wg.wg_info(ifname)
    if not ifaces:
        return
    iface = ifaces[0]
    return iface.get('public_key')


def get_peer_info(ifname, wg, kind=None):
    results = {}
    if kind == 'wireguard' or os.environ.get("SYNTROPY_WIREGUARD"):
        try:
            ss = wg.info(ifname)
        except NetlinkError as e:
            return results
        wg_info = dict(ss[0]['attrs'])
        peers = wg_info.get('WGDEVICE_A_PEERS', [])
        for peer in peers:
            peer = dict(peer['attrs'])
            try:
                results[peer['WGPEER_A_PUBLIC_KEY'].decode('utf-8')] = [allowed_ip['addr'] for allowed_ip in
                                                                        peer['WGPEER_A_ALLOWEDIPS']]
            except KeyError:
                results[peer['WGPEER_A_PUBLIC_KEY'].decode('utf-8')] = []
    else:
        wg = WireGuardRead()
        ifaces = wg.wg_info(ifname)
        if not ifaces:
            return results
        iface = ifaces[0]
        for peer in iface['peers']:
            results[peer['peer']] = peer['allowed_ips']
    return results


def get_peer_info_all(ifname, wg, kind=None):
    results = []
    # TODO NEED FIX pyroute2 wireguard info solution, because of missing peers when getting info.
    # if kind == 'wireguard' or os.environ.get("SYNTROPY_WIREGUARD"):
    #     try:
    #         ss = wg.info(ifname)
    #     except NetlinkError as e:
    #         return results
    #     wg_info = dict(ss[0]['attrs'])
    #     peers = wg_info.get('WGDEVICE_A_PEERS', [])
    #     for peer in peers:
    #         try:
    #             peer_dict = dict(peer['attrs'])
    #             results.append({
    #                 "public_key": peer_dict['WGPEER_A_PUBLIC_KEY'].decode('utf-8'),
    #                 "allowed_ips": [allowed_ip['addr'] for allowed_ip in peer_dict['WGPEER_A_ALLOWEDIPS']],
    #                 "last_handshake": datetime.datetime.strptime(
    #                     peer_dict['WGPEER_A_LAST_HANDSHAKE_TIME']['latest handshake'],
    #                     "%a %b %d %H:%M:%S %Y").isoformat(),
    #                 "keep_alive_interval": peer_dict['WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL'],
    #                 "rx_bytes": peer_dict['WGPEER_A_RX_BYTES'],
    #                 "tx_bytes": peer_dict['WGPEER_A_TX_BYTES'],
    #             })
    #         except KeyError:
    #             continue

    wg = WireGuardRead()
    ifaces = wg.wg_info(ifname)
    if not ifaces:
        return results
    iface = ifaces[0]
    for peer in iface['peers']:
        try:
            results.append({
                "public_key": peer['peer'],
                "last_handshake": datetime.datetime.now().isoformat() if peer['latest_handshake'] else None,
                "keep_alive_interval": int(''.join(filter(str.isdigit, peer.get('persistent_keepalive', '15')))),
                "allowed_ips": peer['allowed_ips'],
            })
        except KeyError:
            continue
    return results


def get_peer_ips(ifname, wg, internal_ip, kind=None):
    peers_info = []
    peers_internal_ip = []
    peers = get_peer_info_all(ifname, wg, kind=kind)
    for peer in peers:
        try:
            peer_internal_ip = next(
                (
                    ip for ip in peer['allowed_ips']
                    if
                    ipaddress.ip_address(ip.split('/')[0]) in ipaddress.ip_network(f"{internal_ip.split('/')[0]}/16",
                                                                                   False)
                ),
                None
            )
        except ValueError:
            continue
        if not peer_internal_ip:
            continue
        peer.update({'internal_ip': peer_internal_ip.split('/')[0]})
        peers_info.append(peer)
        peers_internal_ip.append(peer_internal_ip.split('/')[0])
    return peers_info, peers_internal_ip


def check_if_wireguard_installled():
    return module_loaded('wireguard') or is_tool('wireguard-go')


def check_udp_connection():
    for pings in range(3):
        client_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        client_socket.settimeout(1.0)
        message = b'test'
        addr = ("udp-check.syntropystack.com", 12000)
        client_socket.sendto(message, addr)
        try:
            client_socket.recvfrom(1024)
            return True
        except socket.timeout:
            return False


def ping_internal_ips(ips, count=4, interval=0.5, icmp_id=10000):
    result = {}
    ping_res = multiping(ips, count=count, interval=interval, id=icmp_id)
    for res in ping_res:
        latency_ms = res.avg_rtt if res.is_alive else 5000
        packet_loss = res.packet_loss if res.is_alive else 1
        result[res.address] = get_connection_status(latency_ms, packet_loss)
    return result


def merged_peer_info(wg):
    result = []
    peers_ips = []
    interfaces = read_tmp_file(file_type='iface_info')
    res = {k: v for k, v in interfaces.items() if re.match(WG_NAME_PATTERN, k)}
    for ifname in res.keys():
        if not res[ifname].get('internal_ip'):
            continue
        peer_info, peers_internal_ips = get_peer_ips(ifname, wg, res[ifname]['internal_ip'], kind=res[ifname]['kind'])
        peers_ips += peers_internal_ips
        iface_public_key = get_iface_public_key(ifname)
        if not iface_public_key:
            continue
        result.append(
            {
                "iface": ifname,
                "iface_public_key": iface_public_key,
                "peers": peer_info
            }
        )
    pings = ping_internal_ips(peers_ips, count=1, interval=0.3)
    for iface in result:
        for peer in iface['peers']:
            peer.update(pings[peer['internal_ip']])
    return result



import os
import dns.resolver


def resolve_url_custom_dns(hostname: str, nameservers: list = None):
    if nameservers is None:
        nameservers = ['8.8.8.8']
    my_resolver = dns.resolver.Resolver()

    # 8.8.8.8 is Google's public DNS server
    my_resolver.nameservers = nameservers

    try:
        answer = my_resolver.query(hostname)
        address = list(answer)[0].address
        return f"{address}:8000"
    except dns.resolver.NXDOMAIN:
        return None



from platform_agent.cmd.wg_info import WireGuardRead

import mock


@mock.patch('platform_agent.cmd.wg_info.os.popen')
def test_wireguard_read(patch_cmd_read, mock_wg_show, wg_show_dict):
    patch_cmd_read().read.return_value = mock_wg_show
    wg = WireGuardRead()
    wg_info = wg.wg_info()
    assert wg_info == wg_show_dict



import os
import ipaddress
import json
from pathlib import Path

import yaml

CONFIG_FILE = "/etc/syntropy-agent/config.yaml"

AGENT_PATH = "/etc/syntropy-agent"
AGENT_PATH_TMP = f"{AGENT_PATH}/tmp"

class ConfigException(Exception):
    pass


class Config:

    _data = None

    def __init__(self):
        agent_dir = Path(AGENT_PATH)
        if not agent_dir.is_dir():
            agent_dir.mkdir()
            agent_dir.chmod(0o700)
        tmp_dir = Path(AGENT_PATH_TMP)
        if not tmp_dir.is_dir():
            tmp_dir.mkdir()
            tmp_dir.chmod(0o700)
        if os.environ.get("SYNTROPY_NETWORK_API", '').lower() == "kubernetes" and not os.environ.get('SYNTROPY_AGENT_NAME'):
            try:
                os.environ['SYNTROPY_AGENT_NAME'] = open("/var/run/secrets/kubernetes.io/serviceaccount/namespace").read()
            except FileNotFoundError:
                pass
        if os.environ.get("SYNTROPY_API_KEY"):
            return
        if os.environ.get("SYNTROPY_LOG_LEVEL"):
            if os.environ["SYNTROPY_LOG_LEVEL"].upper() in ['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']:
                os.environ["SYNTROPY_LOG_LEVEL"] = os.environ["SYNTROPY_LOG_LEVEL"].upper()
            elif not (os.environ["SYNTROPY_LOG_LEVEL"].isdigit() and os.environ["SYNTROPY_LOG_LEVEL"] in ['0', '10', '20', '30', '40', '50']):
                os.environ["SYNTROPY_LOG_LEVEL"] = '20'
        else:
            os.environ["SYNTROPY_LOG_LEVEL"] = '20'

        if os.environ.get('SYNTROPY_USER_API') == 'DOCKER' and not os.environ.get('SYNTROPY_DOCKER_URL'):
            raise ConfigException(f"For Docker API, you must provide SYNTROPY_DOCKER_URL")

        config_file = Path(CONFIG_FILE)
        if not config_file.is_file():
            print(f"Config file was not found in {CONFIG_FILE}")
            raise ConfigException(f"Config file was not found in {CONFIG_FILE}")
        env_conf = self.get_config()
        if env_conf.get('name') and type(env_conf['name']) == str:
            os.environ[f"SYNTROPY_AGENT_NAME"] = env_conf['name']
        for k, v in env_conf.get('connection', {}).items():
            if type(v) in [int, str]:
                os.environ[f"SYNTROPY_{k.upper()}"] = str(v)

    @staticmethod
    def get_config():
        try:
            with open(CONFIG_FILE) as f:
                config_dict = yaml.safe_load(f)
                return config_dict
        except FileNotFoundError:
            return {}

    @staticmethod
    def get_list_item(key: str):
        if os.environ.get(f'SYNTROPY_{key.upper()}'):
            result = os.environ.get(f'SYNTROPY_{key.upper()}').split(',')
            return result
        result = Config.get_config().get(key, [])
        if type(result) != list:
            result = []
        return result


    @staticmethod
    def get_valid_allowed_ips():

        def update_results(result: list, subnet: str, subnet_name: str):
            try:
                ip_network = ipaddress.ip_interface(subnet)
            except ValueError:
                return result
            result.append(
                {
                    'agent_network_iface': subnet_name,
                    'agent_network_subnets': [ip_network.with_prefixlen],
                    'agent_network_ports': {'udp': [], 'tcp': []},
                }
            )
            return result

        results = []
        if os.environ.get('SYNTROPY_ALLOWED_IPS'):
            try:
                allowed_ips = json.loads(os.environ['SYNTROPY_ALLOWED_IPS'])
            except json.JSONDecodeError:
                return []
            for allowed_ip in allowed_ips:
                for k, v in allowed_ip.items():
                    if not (type(k) == type(v) == str):
                        continue
                    update_results(results, k, v)
            return results
        allowed_ips = Config.get_config().get('allowed_ips', [])
        for allowed_ip in allowed_ips:
            if allowed_ip.get('name') and allowed_ip.get('subnet'):
                try:
                    ip_network = ipaddress.ip_interface(allowed_ip['subnet'])
                except ValueError:
                    continue
                results = update_results(results, ip_network.with_prefixlen, allowed_ip['name'])
        return results



import json
import logging
import threading
import time
import os

from platform_agent.lib.ctime import now
from pyroute2 import IPDB
from kubernetes import client, config

logger = logging.getLogger()


class KubernetesConfigException(Exception):
    pass


class KubernetesNetworkWatcher(threading.Thread):

    def __init__(self, ws_client):
        current_namespaces = os.environ.get('SYNTROPY_NAMESPACE', None)
        super().__init__()
        try:
            config.load_incluster_config()
            if not current_namespaces:
                current_namespaces = open("/var/run/secrets/kubernetes.io/serviceaccount/namespace").read()

        except config.config_exception.ConfigException:
            try:
                config.load_kube_config()
            except config.config_exception.ConfigException:
                raise KubernetesConfigException("Couldn't find config")
        if current_namespaces:
            self.namespace_list = current_namespaces.split(',')
        else:
            self.namespace_list = []
        logger.debug(f"['KUBERNETES_API'] - Namespace {self.namespace_list}")
        self.v1 = client.CoreV1Api()
        self.ws_client = ws_client
        self.stop_kubernetes_watcher = threading.Event()
        self.interval = 10

        with IPDB() as ipdb:
            self.ifaces = [k for k, v in ipdb.by_name.items() if any(
                substring in k for substring in ['syntropy_'])]
        self.daemon = True

    def run(self):
        ex_result = []
        while not self.stop_kubernetes_watcher.is_set():
            result = []
            if not self.namespace_list:
                return
            for namespace in self.namespace_list:
                ret = self.v1.list_namespaced_service(namespace)
                for i in ret.items:
                    if not i.metadata.name:
                        continue
                    ports = {'udp': [], 'tcp': []}
                    if not i.spec.ports:
                        continue
                    ports['tcp'] = [port.port for port in i.spec.ports if port.protocol == 'TCP']
                    ports['udp'] = [port.port for port in i.spec.ports if port.protocol == 'UDP']
                    result.append(
                        {
                            'agent_service_subnets': f"{i.spec.cluster_ip}/32",
                            'agent_service_name': f"{i.metadata.name}-{namespace}",
                            'agent_service_ports': ports,
                            'agent_service_uptime': i.metadata.creation_timestamp.isoformat(),
                        }
                    )

            status = getattr(self.ws_client.ws, 'sock')

            if result != ex_result and status and status.status:
                self.ws_client.send(json.dumps({
                    'id': "ID." + str(time.time()),
                    'executed_at': now(),
                    'type': 'KUBERNETES_SERVICE_INFO',
                    'data': result
                }))
                ex_result = result
            time.sleep(10)

    def join(self, timeout=None):
        self.stop_kubernetes_watcher.set()
        super().join(timeout)



import time
import socket
import os
import threading

from prometheus_client import start_http_server, Metric, REGISTRY

from platform_agent.cmd.lsmod import module_loaded
from platform_agent.cmd.wg_info import WireGuardRead
from platform_agent.files.tmp_files import get_peer_metadata
from platform_agent.wireguard.helpers import merged_peer_info
from pyroute2 import WireGuard


class JsonCollector(object):
    def __init__(self, interval=10):
        self.interval = interval
        self.wg = WireGuard() if module_loaded("wireguard") else WireGuardRead()

    def collect(self):
        # Fetch the JSON
        peer_info = merged_peer_info(self.wg)
        peer_metadata = get_peer_metadata()
        for iface in peer_info:
            metric = Metric(f"interface_info_{iface['iface']}",
                            'interface_information', 'summary')
            for peer in iface['peers']:
                peer.update(peer_metadata.get(peer['public_key'], {}))
                for k, v in peer.items():
                    if k not in ['latency_ms', 'packet_loss', 'rx_bytes', 'tx_bytes']:
                        continue
                    metric.add_sample(f"iface_information_{k}",
                                      value=str(v),
                                      labels={
                                          'hostname': os.environ.get('SYNTROPY_AGENT_NAME', socket.gethostname()),
                                          'ifname': iface['iface'],
                                          'peer': peer['public_key'],
                                          'internal_ip': peer['internal_ip'],
                                          "device_id": peer.get('device_id'),
                                          "device_name": peer.get('device_name'),
                                          "device_public_ipv4": peer.get('device_public_ipv4')
                                      })
            yield metric


class  NetworkExporter(threading.Thread):

    def __init__(self, port=18001):
        super().__init__()
        self.stop_network_exporter = threading.Event()
        self.exporter_port = port
        self.daemon = True

    def run(self):
        start_http_server(self.exporter_port)
        REGISTRY.register(JsonCollector())
        while self.stop_network_exporter.is_set(): time.sleep(1)

    def join(self, timeout=None):
        self.stop_network_exporter.set()
        super().join(timeout)



from pathlib import Path

RT_TABLE_PATH = "/etc/iproute2/rt_tables"


def read_rt_table():
    path = "/etc/iproute2/rt_tables"
    rt_table = Path(path)
    res = rt_table.read_text()
    rt_table_dict = {}
    for line in res.splitlines():
        if line[0] == "#":
            continue
        line = line.strip()
        line.replace('\t', ' ')
        res = line.split()
        if len(res) >= 2 and res[0].isdigit():
            rt_table_dict[res[0]] = {'name': res[1]}
    return rt_table_dict


def insert_rt_table(table_id, name):
    path = "/etc/iproute2/rt_tables"
    with open(path, 'a') as rt_table:
        rt_table.write(f'{table_id} {name}\n')


def delete_rt_table(table_id, name):
    with open(RT_TABLE_PATH, "r") as f:
        lines = f.readlines()
    with open(RT_TABLE_PATH, "w") as f:
        for line in lines:
            if line != f'{table_id} {name}\n':
                f.write(line)


def get_available_rt_table():
    table = read_rt_table()
    for i in range(1, 252):
        if table.get(str(i)) is None:
            return str(i)



import logging
import socket
import subprocess

from pyroute2 import IPRoute, NetlinkError
from ipaddress import IPv4Network, ip_network

from platform_agent.files.tmp_files import get_agent_id_by_text

logger = logging.getLogger()


class Routes:
    def __init__(self):
        self.ip_route = IPRoute()

    def ip_route_add(self, ifname, ip_list, gw_ipv4):
        devices = self.ip_route.link_lookup(ifname=ifname)
        dev = devices[0]
        statuses = []
        routes = self.ip_route.get_routes(family=socket.AF_INET)
        network_list = []
        for route in routes:
            try:
                ip_cidr = (dict(route['attrs'])['RTA_DST'] + '/' + str(route['dst_len']))
                network_list.append(IPv4Network(ip_cidr))
            except (KeyError, ValueError):
                continue
        for ip in ip_list:
            # format ip to strict format
            formatted_ip = ip_network(ip, False)
            result = {'ip': ip, 'status': 'OK'}
            for network in network_list:
                if formatted_ip == network:
                    continue
                elif formatted_ip.overlaps(network):
                    agent_id = get_agent_id_by_text(ip)
                    result = {
                        'ip': ip,
                        'status': 'ERROR',
                        'msg': f'Service ip: {ip} intersects agent_id: {agent_id} CIDR {network}'
                    }
                    break
            if result['status'] == 'ERROR':
                statuses.append(result)
                continue
            try:
                self.ip_route.route('add', dst=formatted_ip.with_prefixlen, gateway=gw_ipv4, oif=dev)
            except NetlinkError as error:
                if error.code != 17:
                    result.update({'status': "ERROR", 'msg': str(error)})
                elif dict(self.ip_route.get_routes(dst=formatted_ip.with_prefixlen)[0]['attrs']).get('RTA_OIF') != dev:

                    logger.debug(f"[WG_CONF] add route failed [{ip}] - already exists")
                    result.update({'status': "ERROR", 'msg': "OVERLAP"})
            statuses.append(result)
        return statuses

    def ip_route_replace(self, ifname, ip_list, gw_ipv4):
        devices = self.ip_route.link_lookup(ifname=ifname)
        dev = devices[0]
        for ip in ip_list:
            try:
                self.ip_route.route('replace', dst=ip, gateway=gw_ipv4)
            except NetlinkError as error:
                if error.code != 17:
                    raise

    def ip_route_del(self, ifname, ip_list, scope=None):
        devices = self.ip_route.link_lookup(ifname=ifname)
        dev = devices[0]
        for ip in ip_list:
            try:
                self.ip_route.route('del', dst=ip, oif=dev, scope=scope)
            except NetlinkError as error:
                if error.code not in [17, 3, 19]:
                    raise

    def create_rule(self, internal_ip, rt_table_id):
        self.ip_route.flush_rules(table=rt_table_id)
        self.ip_route.flush_routes(table=rt_table_id)
        self.ip_route.rule('add', src=internal_ip, table=rt_table_id)

    def clear_unused_routes(self, ifname, ips):
        devices = self.ip_route.link_lookup(ifname=ifname)
        if len(devices) > 0:
            dev = devices[0]
        else:
            return
        routes = self.ip_route.get_routes(family=socket.AF_INET)
        already_used_ips = []
        for route in routes:
            try:
                attrs = dict(route['attrs'])
                if attrs.get('RTA_OIF') == dev and route.get('type') == 1:
                    already_used_ips.append(attrs.get('RTA_DST') + '/' + str(route['dst_len']))
            except (KeyError, ValueError):
                continue
        remove_ips = set(already_used_ips) - set(ips)
        self.ip_route_del(ifname, remove_ips)

    def clear_unused_iface_addrs(self, ifname, current_addr):
        addrs = self.ip_route.get_addr(label=ifname)
        for addr in addrs:
            if current_addr.split('/')[0] != str(dict(addr.get('attrs')).get('IFA_ADDRESS')):
                subprocess.run(
                    ['ip', 'addr', 'del', f"{dict(addr.get('attrs')).get('IFA_ADDRESS')}", 'dev', ifname],
                    check=False, stderr=subprocess.DEVNULL
                )


import os
import logging
import socket
import requests

import docker
from requests.exceptions import ConnectionError, SSLError
from urllib3.exceptions import ProtocolError, NewConnectionError

from platform_agent.docker_api.helpers import format_networks_result, format_container_result
from platform_agent.config.settings import Config

logger = logging.getLogger()


def get_ip_addr():
    try:
        resp = requests.get("https://ip.syntropystack.com/")
        return {
            "external_ip": resp.json()
        }
    except (NewConnectionError, SSLError, ConnectionError):
        return {}


def get_location():
    return {
        "location_lat": os.environ.get('SYNTROPY_LAT'),
        "location_lon": os.environ.get('SYNTROPY_LON'),
    }


def get_network_info():
    network_info = []
    if os.environ.get("SYNTROPY_NETWORK_API", '').lower() == "docker":
        try:
            docker_client = docker.from_env()
            networks = docker_client.networks()
            network_info = format_networks_result(networks)
        except (ProtocolError, ConnectionError):
            network_info = []
    network_info.extend(Config.get_valid_allowed_ips())
    return {
        "network_info": network_info
    }


def get_container_results():
    container_info = []
    if os.environ.get("SYNTROPY_NETWORK_API", '').lower() == "docker":
        try:
            docker_client = docker.from_env()
            networks = docker_client.containers()
            container_info = format_container_result(networks)
        except (ProtocolError, ConnectionError):
            container_info = []
    return {
        "container_info": container_info
    }


def get_srevice_status():
    if os.environ.get('SYNTROPY_SERVICES_STATUS') and os.environ.get('SYNTROPY_SERVICES_STATUS').lower() == 'true':
        return True
    else:
        return False


def get_provider():
    try:
        return int(os.environ['SYNTROPY_PROVIDER'])
    except (ValueError, KeyError):
        return None


def get_tags():
    tags = []
    for tag in Config.get_list_item('tags'):
        if len(tag) < 3:
            continue
        tags.append(tag)
    return tags


def get_info():
    return {
        "agent_name": os.environ.get('SYNTROPY_AGENT_NAME', socket.gethostname()),
        "agent_provider": get_provider(),
        "agent_category": os.environ.get('SYNTROPY_CATEGORY', None),
        "service_status": get_srevice_status(),
        "agent_tags": get_tags(),
        "network_ids": Config.get_list_item('network_ids'),
    }


def gather_initial_info():
    Config()
    result = {}
    result.update(get_ip_addr())
    result.update(get_network_info())
    result.update(get_info())
    result.update(get_container_results())
    result.update(get_location())
    return result



import json
import logging
import threading
import time
import queue
import traceback

from platform_agent.files.tmp_files import update_tmp_config_dump
from platform_agent.lib.ctime import now
from platform_agent.wireguard import WgConfException, WgConf

logger = logging.getLogger()


class WgExecutor(threading.Thread):
    CMD_TYPE = "WG_CONF"

    def __init__(self, client):
        super().__init__()
        self.queue = queue.Queue()
        self.client = client
        self.stop_wg_executor = threading.Event()
        self.wg = None
        self.wgconf = WgConf(client)
        self.daemon = True

        threading.Thread.__init__(self)

    def get_from_queue(self):
        payloads = {}
        t_end = time.time() + 0.2  # run for 1 second
        while time.time() < t_end:
            try:
                message = self.queue.get(timeout=0.1)
            except queue.Empty:
                continue
            request_id = message['request_id']
            payloads[request_id] = []
            data = message['data'] if type(message['data']) == list else [message['data']]
            for payload in data:
                try:
                    payloads[request_id].append(
                        {
                            "fn_name": payload['fn'],
                            "fn_args": payload['args'],
                            "request_id": request_id
                        }
                    )
                    update_tmp_config_dump(payload)
                except AttributeError as e:
                    logger.warning(e)
                    payloads[request_id].append(
                        {"request_id": request_id, "error": f"{e}"})
        return payloads

    def run(self):
        while True:
            payloads = self.get_from_queue()
            if not payloads:
                continue
            logger.debug(f"[WG_EXECUTOR] - Received {payloads}")
            for request_id in list(payloads.keys()):
                try:
                    self.execute_payload(request_id, payloads)
                except:  # noqa Catch all errors and report to controller
                    # Catch all exceptions that not handled
                    logger.debug(f"[WG_EXECUTOR] - catched error")
                    self.send_error(request_id)

    def execute_payload(self, request_id, payloads):
        result = {}
        ok = {}
        errors = []
        for payload in payloads[request_id]:
            if payload.get('error'):
                errors.append(payload['error'])
            else:
                try:
                    fn = getattr(self.wgconf, payload['fn_name'])
                    result = fn(**payload['fn_args'])
                    ok.update(
                        {"fn": payload['fn_name'], "data": result, "args": payload['fn_args']})
                except WgConfException as e:
                    logger.error(f"[WG_EXECUTOR] failed. exception = {str(e)}, data = {payload}")
                    errors.append({payload['fn_name']: str(e), "args": payload['fn_args']})
            logger.debug(f"[WG_EXECUTOR] - Results {result}")
            response = {
                'id': request_id,
                'executed_at': now(),
                'type': self.CMD_TYPE,
            }
            if errors:
                response.update({'error': errors, 'data': {}})
            elif ok:
                response.update({'data': ok})

            self.client.send(json.dumps(response))

    def join(self, timeout=None):
        self.stop_wg_executor.set()
        super().join(timeout)

    def send_error(self, request_id):
        traceback.print_exc()
        result = {
            'error': {
                'traceback': traceback.format_exc(),
            }
        }
        logger.error(result)
        if result:
            payload = {
                'id': request_id,
                'executed_at': now(),
                'type': self.CMD_TYPE,
            }
            if isinstance(result, dict) and ('error' in result):
                payload.update(result)
            else:
                payload.update({'data': result})
            payload = json.dumps(payload)
            self.client.send(payload)



from setuptools import setup, find_packages, Extension

import os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

if os.environ.get("CI_SYNTROPY_VERSION"):
    version = os.environ["CI_SYNTROPY_VERSION"]
else:
    # Development version of the package
    version = "devel"

setup(
    name="syntropy-agent",
    version=version,
    py_modules=['platform-agent'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        'requests==2.24.0',
        'urllib3<1.26',
        'pyroute2==0.5.12',
        'websocket-client==0.57.0',
        'PyNaCl==1.3.0',
        'docker-py==1.10.6',
        'icmplibv2==1.0.6',
        'PyYAML==5.3.1',
        'dnspython==1.16.0',
        'iperf3==0.1.11',
        'prometheus-client==0.8.0',
        'psutil==5.7.2',
        'kubernetes==11.0.0',
    ],
    packages=find_packages(),
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    entry_points={
        'console_scripts': [
            'syntropy_agent = platform_agent.__main__:main'
        ]
    },
)



from platform_agent.agent_api import AgentApi

import mock

SINGLE_CALL = 1


@mock.patch('platform_agent.agent_api.gather_initial_info')
def test_call(patch_gather_initial_info, gather_initial_info_payload, GET_INFO, request_id):
    agent_api = AgentApi(mock.MagicMock(), prod_mode=False)
    agent_api.call(GET_INFO, gather_initial_info_payload, request_id)
    assert patch_gather_initial_info.called
    assert patch_gather_initial_info.call_count == SINGLE_CALL
    assert patch_gather_initial_info.call_args[0] == ()


@mock.patch('platform_agent.agent_api.gather_initial_info')
def test_call_bad(patch_gather_initial_info, gather_initial_info_payload_bad, GET_INFO, request_id):
    agent_api = AgentApi(mock.MagicMock(), prod_mode=False)
    agent_api.call(GET_INFO, gather_initial_info_payload_bad, request_id)
    assert not patch_gather_initial_info.called
    assert patch_gather_initial_info.call_count == 0


@mock.patch('platform_agent.agent_api.gather_initial_info')
def test_call_error(patch_gather_initial_info, gather_initial_info_payload, GET_INFO, request_id):
    agent_api = AgentApi(mock.MagicMock(), prod_mode=False)
    patch_gather_initial_info.return_value = {'error': 'SOME_ERROR'}
    result = agent_api.call(GET_INFO, gather_initial_info_payload, request_id)
    assert result == {'error': 'SOME_ERROR'}
    assert patch_gather_initial_info.called
    assert patch_gather_initial_info.call_count == SINGLE_CALL
    assert patch_gather_initial_info.call_args[0] == ()


@mock.patch('platform_agent.agent_api.json.dumps')
@mock.patch('platform_agent.agent_api.update_tmp_file')
@mock.patch('platform_agent.agent_api.WgConf.clear_peers')
@mock.patch('platform_agent.agent_api.WgConf.clear_interfaces')
@mock.patch('platform_agent.agent_api.WgConf.create_interface')
def test_config_info(patch_create_interface, patch_clear_interfaces, patch_clear_peers, patch_update_tmp_file, patch_json_dumps, config_info_int_check, CONFIG_INFO, request_id):
    agent_api = AgentApi(mock.MagicMock(), prod_mode=False)
    agent_api.call(CONFIG_INFO, config_info_int_check, request_id)
    assert patch_clear_interfaces.called
    assert patch_clear_peers.called
    assert patch_create_interface.call_count == len(config_info_int_check['vpn'])
    assert patch_clear_interfaces.call_args[0][0] == config_info_int_check['vpn']


@mock.patch('platform_agent.agent_api.WireguardPeerWatcher')
def test_WG_INFO(patch_WireguardPeerWatcher, wg_info_payload, WG_INFO, request_id):
    agent_api = AgentApi(mock.MagicMock(), prod_mode=False)
    patch_WireguardPeerWatcher.return_value = mock.MagicMock()
    result = agent_api.call(WG_INFO, wg_info_payload, request_id)
    assert not result
    assert patch_WireguardPeerWatcher.called
    assert patch_WireguardPeerWatcher.call_count == SINGLE_CALL
